/* Copyright (c) 1996 University of Utah.  All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for NON-COMMERCIAL purposes and without fee is
 * hereby granted provided that this copyright notice appears in all
 * copies.
 *
 * See http://www.cs.utah.edu/~hamlet for more educational resources.
 *
 * The University of Utah MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT
 * THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT. The University of Utah SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package Cannon;

import java.awt.*;
import Cannon.*;
import hamlet.*;
import java.util.Vector;

//
//  Class cannonPanel (containing one (1) cannon)
//
public class cannonPanel extends Panel implements Runnable {
	
  public static int offsetBottom = 27;
  public static int offsetTop = 0;
  public static int offsetLeft = 35;
  public static int offsetRight = 0;
  
  public cannonControls controlPanel;
  public Thread animatorThread;
  
  int velocity, angle, height, range, scalingRatio = 2, currentX, currentY;
  int distanceAtImpact;
  double timeAtImpact;
  double elapsed = 0.0, g;
  
  // spiffy new labeling system
  NumLabel xPos, yPos, cTime;
  FontMetrics fm;
    
  Vector curves = new Vector(2, 5);
  Vector thisCurve;

  // double buffering scheme
  Dimension offDimension;
  Image offImage;
  Image oldImage;
  Graphics offGraphics;

    
public cannonPanel()  {
    animatorThread = null;
    setBackground( Color.white );
}
      
      
public void reset() {
   // System.out.print("RESET ");
    makeCleanGraphics();
    curves.removeAllElements();
    xPos = null;
    yPos = null;
    cTime = null;
    halt();
    update( getGraphics() );
}


public void halt() {
   // System.out.println("HALT ");
    update( getGraphics() );
    animatorThread = null;
}
	
	
public synchronized void update( Graphics g )  {
  //  System.out.print("UPDATE:  ");

	g.drawImage( offImage, 0, 0, this );	
} 


public synchronized void paint ( Graphics g )  {
 	if ( (offImage==null) || (offGraphics==null) ) {
 	 	makeCleanGraphics();
    }
	update( g );
}

public synchronized void drawFrame( Graphics g )  {
  		int distance = getNewX() / scalingRatio;
		int altitude = (getNewY() / scalingRatio)+1;
		int xOnCanvas, yOnCanvas;
		double elap;

		if (getNewY() >= 0) {
		  xPos.dvalue = getNewX();
		  yPos.dvalue = getNewY();
		  elap = elapsed;
		}
		else {
		  xPos.dvalue = distanceAtImpact;
		  yPos.dvalue = 0;
		  elap = timeAtImpact;
		  distance = distanceAtImpact / scalingRatio;
		  altitude = 0;
		}
		elap = elap * 10;
		elap = Math.floor(elap);
		elap = elap * .1;
		cTime.dvalue = elap ;

		/* paint on the numbers */
		xPos.paint( offGraphics );
		yPos.paint( offGraphics );
		cTime.paint( offGraphics );

		xOnCanvas = distance;
		yOnCanvas = altitude;

		// this ensures that we only draw onscreen
		if ( altitude < bounds().height-offsetBottom ||
		    ( distance < bounds().width-offsetLeft ) )  {
			//g.clipRect( offsetLeft+distance-2, 
		        //            bounds().height-offsetBottom-altitude-2, 
			//			6, 6 );
			//System.out.println(" Xpos: "+(offsetLeft+distance-5)+
			//		"    Ypos: " +(bounds().height-offsetBottom-altitude-5));
			//g.drawImage( offImage, 0, 0, this );
			offGraphics.setColor( Color.black );
			offGraphics.drawLine( offsetLeft+currentX, bounds().height-offsetBottom-currentY, 
						offsetLeft+distance, bounds().height-offsetBottom-altitude );
			currentX = distance;
			currentY = altitude;
		}
		g.drawImage( offImage, 0, 0, this );
		thisCurve.addElement(new Coord(xOnCanvas, yOnCanvas));
/*   		if ( ( altitude <= 0 ) || ((distance+offsetLeft) > bounds().width ) ) {*/
   		if ( altitude <= 0 )  {
	  		// System.out.println("distance ( " + distance + " ) > range( " 
	  		// 					+ range + " ) " );
	  		// System.out.println("OR altitude ( " + altitude + " ) <= 0 " );
	  		// g.clipRect( 0,0, size().width, size().height );
	  		// g.clearRect( offsetLeft, bounds().height-offsetBottom, 
	  		//			 bounds().width-offsetLeft, 30 );
	  	  	postEvent( new Event( new Button(), Event.ACTION_EVENT, "Halt" ) );
	  	} 
} 
    

public synchronized void go() {
    animatorThread = new Thread( this );
    
    //  set the instance variables for this trajectory
    velocity = controlPanel.velocity.getIntText();
    /*    velocity = retrieveText( controlPanel.velocity, 0, 150 );*/
    angle = controlPanel.gunAngle.getIntText();
    /*    angle = retrieveText( controlPanel.gunAngle, 0, 90 );*/
    g = controlPanel.gravity.getDoubleText();
      /*    try{
	g = (Double.valueOf( controlPanel.gravity.getText() )).doubleValue();
    } catch( NumberFormatException e ) {
	controlPanel.gravity.setText("9.8");
	g = 9.8;
    }*/
    height = (int)Math.round( (velocity * velocity * 
			       Math.sin( radians(angle) ) *
			       Math.sin( radians(angle) ))
			      / (2 * g) );
    range = (int)Math.round( (velocity * velocity *
			      Math.sin( radians(2*angle) )) / g );
    //System.out.println("UNSCALED:  Range = "+range+"     Height = "+height );
    // SCALE!!
    height /= scalingRatio;
    range /= scalingRatio;
    currentX = 0;
    currentY = 0;
    timeAtImpact = (velocity * Math.sin( radians(angle) ) + 
                   Math.sqrt( velocity * Math.sin( radians(angle) ) *
			      velocity * Math.sin( radians(angle) ) + 2*g))/g ;  
    distanceAtImpact = (int)Math.round 
                    (velocity * timeAtImpact * Math.cos( radians(angle)));
    animatorThread.start();
} 
 

public void run()  {
    thisCurve = new Vector(50, 50);  // Storage for the current curve in 
	  			     // case we need to redraw it.
    animatorThread.setPriority(Thread.MAX_PRIORITY);
    long startTime = System.currentTimeMillis();
    elapsed = 0;
    Graphics g = getGraphics();
    
    xPos = new NumLabel( this, 0, "m", offsetLeft+20+fm.stringWidth("Distance: "), 
    					offsetTop+25-fm.getAscent(), Label.LEFT );
    xPos.paint( offGraphics );

    yPos = new NumLabel( this, 0, "m", offsetLeft+20+fm.stringWidth("Altitude: "),
     					offsetTop+45-fm.getAscent(), Label.LEFT );
    yPos.paint( offGraphics );
    
    cTime = new NumLabel( this, 0, "s", offsetLeft+20+fm.stringWidth("Altitude: "),
     					offsetTop+65-fm.getAscent(), Label.LEFT );
    cTime.paint( offGraphics );

    /*new NumLabel( this, 0, "m", offsetLeft+20+fm.stringWidth("Time: "),
			                  offsetTop+65-fm.getAscent(), Label.LEFT )*/
    //  let's animate
    while ( (Thread.currentThread() == animatorThread) && (animatorThread != null) ) {
		elapsed = ((double)(System.currentTimeMillis() - startTime) / 
				   (double) 1000.0);
		//System.out.println("elapsed time = " + elapsed );
		drawFrame( g );
		getToolkit().sync();
    }
    curves.addElement(thisCurve);
    thisCurve = null;
}
    

//  returns the position in the x dimension, relative to some origin
int getNewX() {
	return (int)Math.round( velocity * elapsed
				* Math.cos ( radians( angle ) ) );
    }
    
    
//  returns the ball's position in the y dimension, relative to some origin
int getNewY()  {
    int alt;
    alt = (int)Math.round( velocity * elapsed * Math.sin( radians( angle ) ) -
			   ( .5 * g * elapsed * elapsed ) );
    return alt;
  }


void createGrid( Graphics g ) {
    int i, interval = 25;
    String value;
    g.setFont( getGraphics().getFont() );
    
    g.setColor( Color.gray );
    g.fillRect(  0,0,size().width,size().height );
    g.setColor( Color.white );
    g.fillRect( offsetLeft, offsetTop, 
		size().width-offsetRight-offsetLeft,
		size().height-offsetBottom-offsetTop );
    g.setColor( Color.pink.darker() );
    
    /*  horizontal lines, left-side numbers  */
    g.drawLine( offsetLeft, bounds().height-offsetBottom, 
	       bounds().width, bounds().height-offsetBottom );
    for (i = bounds().height-offsetBottom-interval; i > 0; i-=interval) {
	g.drawLine( offsetLeft, i, 
		    bounds().width, i );
	g.setColor( Color.black );
	value =
	    Integer.toString( (bounds().height-offsetBottom-i)*scalingRatio );
	g.drawString( value, 5, i+(fm.getAscent()/2) );
	g.setColor( Color.pink.darker() );
    }
    
    /*  vertical lines, bottom numbers */
    interval = 40;
    g.drawLine( 0+offsetLeft, 0, 0+offsetLeft, bounds().height-offsetBottom );
    for (i = interval; i < bounds().width-interval; i+=interval) {
	//System.out.println("at i = "+i+"  we draw" );
	g.drawLine( i+offsetLeft, 0, i+offsetLeft,
		    bounds().height-offsetBottom );
	g.setColor( Color.black );
	value = Integer.toString( i*scalingRatio );
	g.drawString( value, i+offsetLeft-( fm.stringWidth(value)/2 ), 
		      bounds().height-offsetBottom+20 );
	g.setColor( Color.pink.darker() );
    }
	g.setColor( Color.black );
	g.drawString( "Distance: ", offsetLeft+20, offsetTop+25 );
	g.drawString( "Altitude: ", offsetLeft+20, offsetTop+45 );
        g.drawString( "Time: ",     offsetLeft+20, offsetTop+65 );
}
   
    
int retrieveText( TextField tf, int min, int max ) {
	String  s;
	int average, value;
		
	s = tf.getText();
	average = (int)( (max-min)/2 );
	try {
	   value = Integer.parseInt(s);
	} catch (Exception e) {
	    value = average;
	}
	if ( value > max ) value = max;
	if ( value < min ) value = min;
	tf.setText( Integer.toString(value) );
	return value;
}


double radians( int angle ) {
	return ( ( Math.PI * angle ) / 180.0 );
}


void makeCleanGraphics() {
    offImage = createImage( size().width, size().height );
    offGraphics = offImage.getGraphics();
    fm = getFontMetrics( getFont() );
    createGrid( offGraphics );
}

void makeCleanGraphics(boolean newSize) {
    int i, j;
    Coord prevPnt, curPnt;
    Vector curCurve;

    offImage = createImage( size().width, size().height );
    offGraphics = offImage.getGraphics();
    fm = getFontMetrics( getFont() );
    createGrid( offGraphics );
    /* If there are any curves which need to be redrawn. */
    if (!curves.isEmpty() || thisCurve != null) {
      for ( i = 0; i < curves.size(); i++) {
	prevPnt = new Coord(0,0);
	curCurve = (Vector) curves.elementAt(i);
	for (j = 0; j < curCurve.size(); j++) {
	  curPnt = (Coord) curCurve.elementAt(j);
	  offGraphics.drawLine(prevPnt.x+offsetLeft, bounds().height-offsetBottom-prevPnt.y,
			       curPnt.x+offsetLeft, bounds().height-offsetBottom-curPnt.y);
	  prevPnt = curPnt;
	}
      } /* End of for curves.size() */
      if (thisCurve != null) {
	prevPnt = new Coord(0,0);
	curCurve = thisCurve;
	for (j = 0; j < curCurve.size(); j++) {
	  curPnt = (Coord) curCurve.elementAt(j);
	  offGraphics.drawLine(prevPnt.x+offsetLeft, bounds().height-offsetBottom-prevPnt.y,
			       curPnt.x+offsetLeft, bounds().height-offsetBottom-curPnt.y);
	  prevPnt = curPnt;
	}
      } /* End of theCurve != null */
    } /* End of curves not empty of the Curve != null */
    if (xPos != null) {
      xPos.paint( offGraphics);
      yPos.paint( offGraphics);
      cTime.paint( offGraphics);
    }
    /*offGraphics.drawImage( oldImage, 0, 0, this);*/
}

public void resize (int w, int h) {
    //System.out.println(" resize (int w, int h) is sending " + h );
   if (offImage != null) {
     oldImage = createImage( size().width, size().height);
     (oldImage.getGraphics()).drawImage(offImage, 0, 0, this);
   }
    super.resize( w, h );
    adjustSize();
}
    
    
public void resize (Dimension d) {
    //System.out.println(" resize (Dimension d) is sending " + d.height );
   if (offImage != null) {
     oldImage = createImage( size().width, size().height);
     (oldImage.getGraphics()).drawImage(offImage, 0, 0, this);
   }
    super.resize( d );
    adjustSize();
}
    
    
public void reshape (int x, int y, int w, int h) {
   /*if (offImage != null) {
     oldImage = createImage( size().width, size().height);
     (oldImage.getGraphics()).drawImage(offImage, 0, 0, this);
   }*/
   super.reshape( x,y,w,h );
   //System.out.println(" reshape is sending " + h );
   adjustSize();
}
    
    
public synchronized void adjustSize () {
    try {
      makeCleanGraphics(true);
      update( getGraphics() );
      getToolkit().sync();
    }  catch( Exception e)  {
      validate();
      // System.out.println("caught an exception in adjustSize");
      repaint();
    }
}


}  /*  end class cannonPanel */


/* This class is used to represent one plot coordinate when 
 * redrawing the plot */
class Coord extends Object {
  int x;
  int y;
  
  public Coord(int inX, int inY) {
    x = inX;
    y = inY;
  }

}
